/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * Simple example of two processes communicating via buffers.
 *   Ben Moszkowski.  
 *
 * Process 1 is passed a number and iteratively multiples it by 3.
 * Process 2 awaits the result and then iteratively multiplies it by 5.
 *
 */

define process1(n,Buf1,Buf2) = process {
    exists I: {
	{
	    I=0 and Buf1=[]
	    and for 3 times do {
		skip and I:=I+n
	    }
	    and stable Buf1
	};
	{
	    skip and Buf1:=[I] and stable I
	}
    };
    {
	halt(|Buf2|=1) and stable Buf1
    }
}.


define process2(Buf1,Buf2) = process {
    {
	halt(|Buf1|=1) and Buf2=[] and stable Buf2
    };
    exists I: {
	{
	    I=0
	    and for 5 times do {
		skip and I:=I+Buf1[0]
	    }
	    and stable Buf2
	};
	{
	    skip and Buf2:=[I] and stable I
	}
    }
}.


/* run */ define buffer_test() = exists Buf1,Buf2: {
    List(Buf1,1) and List(Buf2,1) and
    halt(|Buf2|=1)
    and format("The answer should be 4*3*5=60.\n")
    and process1(4,Buf1,Buf2)
    and process2(Buf1,Buf2)
    and always format("Buffer1=%8t Buffer2=%8t\n",Buf1,Buf2)
}.

