/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: mem.c,v 2.12 2009/12/07 11:18:30 acau Exp $
 */


#include "system.h"





/*******************************\
*                               *
*  Memory management routines   *
*                               *
\*******************************/

/* new_loc:  initialise a memory cell. */
/*
NODEPTR new_loc (loc,attrib)
register NODEPTR loc;
int attrib;
{
    unsigned i;
    
    for (i=0; i <= 4; i = i + 1) {
        node_mem_memno(loc+i) = i;
        node_mark(loc+i) = F;
        node_mem_isstatic(loc+i) = (attrib == STATIC);
        node_mem_isindirect(loc+i) = F;
        node_mem_time(loc+i) = UNDEF;
        node_mem_isundefined(loc+i) = 0;
        node_mem_hasdefault(loc+i) = 0;
        node_next(loc+i) = NULL;
    }
    printf("new_loc: \n");
    
    return (loc);
}
*/

NODEPTR new_loc1 (loc,attrib,history)
register NODEPTR loc;
int attrib;
unsigned int history;
{
    unsigned int i;
    unsigned int up=3+history;
    
    for (i=0; i <= up; i = i + 1) {
        node_mem_memno(loc+i) = i;
        node_mark(loc+i) = F;
        node_mem_isstatic(loc+i) = (attrib == STATIC);
        node_mem_isindirect(loc+i) = F;
        node_mem_time(loc+i) = UNDEF;
        node_mem_isundefined(loc+i) = 0;
        node_mem_hasdefault(loc+i) = 0;
        node_mem_history(loc+i) = history;
        node_next(loc+i) = NULL;
    }
        /*if (DebugLevel>2) 
          printf("new_loc1: %p, history=%d\n",loc,history);*/
    
    return (loc);
}

/*
NODEPTR mem_block(attrib,size)
int attrib;
register unsigned int size;
{
    NODEPTR blk;
    register NODEPTR loc;

    blk = allocate_mem (size);
    loc = node_blk_base(blk);
    while (size-- > 0) {
        new_loc (loc, attrib);
        loc += offset (1);
    }
    return (blk);
}
*/

NODEPTR mem_block1(attrib,size,history)
int attrib;
register unsigned int size;
unsigned int history;
{
    NODEPTR blk;
    register NODEPTR loc;

    blk = allocate_mem1 (size,history);
    loc = node_blk_base(blk);
    while (size-- > 0) {
        new_loc1 (loc, attrib, history);
        loc += offset2 (1, history);
    }
    return (blk);
}

/*
NODEPTR allocate_mem(size)
unsigned int size;
{
    NODEPTR base_loc;

    if (size == 0)
        return (block_node (0, NULL));
    base_loc = nodep(allocate (offset (size), sizeof (NODE)));
    LocalMemUsed += size;
    return (block_node (size, base_loc));
}
*/

NODEPTR allocate_mem1(size,history)
unsigned int size;
unsigned int history;
{
    NODEPTR base_loc;

    if (size == 0)
        return (block_node (0, NULL));
    base_loc = nodep(allocate (offset2 (size,history), sizeof (NODE)));
    LocalMemUsed += size;
    return (block_node (size, base_loc));
}

/*
int mem_defined(loc)
register NODEPTR loc;
{
    unsigned mem_def;
    
    dereference (loc);
    if (node_mem_isstatic(loc)){
        mem_def = (node_mem_time(loc) != UNDEF);
    }
    else {
        if (is_next (loc) || is_prevnext (loc)) {
            mem_def = (node_mem_time(loc) == Clock + 1);
        }
        else {
            if (is_curr (loc) || is_prev (loc)) {
                mem_def = (node_mem_time(loc) == Clock); 
            }
            else {
                if (StateNumber == 0) {
                    mem_def = F;
                }
                else {
                    mem_def = (node_mem_time(loc) == Clock - 1);  
                }
            }
        }
    }
    
        /*
    if (DebugLevel>2) {
        printf("<CLock=%3d, time=%3d, C=%1d, N=%1d, P=%1d, P2=%1d, MD=%1d>\n",Clock,node_mem_time(loc), is_curr (loc), is_next (loc), is_prev (loc),is_prevnext (loc),mem_def);
        }
    return (mem_def);
}
*/

int mem_defined1(loc)
register NODEPTR loc;
{
    unsigned mem_def;
    unsigned int i;
    
    
    dereference (loc);
    if (node_mem_isstatic(loc)){
        mem_def = (node_mem_time(loc) != UNDEF) && (node_mem_time(loc)<=Clock +1);
    }
    else {
        if (is_next (loc) || is_prev1 (loc,1)) {
            mem_def = (node_mem_time(loc) == Clock + 1);
        }
        else {
            if (is_curr (loc) || is_prev1 (loc,0)) {
                mem_def = (node_mem_time(loc) == Clock); 
            }
            else { mem_def = 0;
                for (i=0; i<node_mem_history(loc); i = i +1 ) {
                    if (is_prev1 (loc,2+i)) {
                            /*printf("mem_def: history=%d, i=%i\n ",node_mem_history(loc),i);*/
                        if (node_mem_time(loc)==UNDEF)
                            mem_def = 0;
                        else {
                            mem_def = (node_mem_time(loc) == Clock - (node_mem_history(loc) - i));
                                /*
                                  printf("mem_def: time=%d clock=%d MD=%1d\n",node_mem_time(loc),Clock,mem_def);*/
                        }
                    }
                }
            }
        }
    }
       
        /*    
    if (DebugLevel>2) {
        printf("<CLock=%3d, time=%3d, MD=%1d>\n",Clock,node_mem_time(loc),mem_def);
    }
        */
    return (mem_def);
}


NODEPTR get_mem (loc)
register NODEPTR loc;
{
    dereference (loc);
    return (node_next(loc));
}

NODEPTR curr_list(list)
NODEPTR list;
{
        /*printf("curr_list: base_loc=%p\n",node_lst_base(list));*/
    return ((is_next (node_lst_size(list)) && is_next (node_lst_base(list))) ?
            list_node (node_lst_blk(list),
                        curr_loc (list_size (list)),
                        curr_loc (list_base (list))) :
             list);
}

NODEPTR next_list(list)
NODEPTR list;
{
        /*printf("next_list: base_loc=%p\n",node_lst_base(list));

    if (is_curr (node_lst_size(list))) {
        printf("is_curr size\n");
    }
    if (is_curr (node_lst_base(list))) {
        printf("is_curr base\n");
        }*/
    
    return ((is_curr (node_lst_size(list)) && is_curr (node_lst_base(list))) ?
            list_node (node_lst_blk(list),
                        next_loc (list_size (list)),
                        next_loc (list_base (list))) :
             list);
}

/*
NODEPTR prev_list(list)
NODEPTR list;
{
    return ((is_prev1 (node_lst_size(list),0) || is_prev1 (node_lst_base(list),0)) ?
            list_node (node_lst_blk(list),
                       prev_loc1 (list_size (list),0),
                       prev_loc1 (list_base (list)),0) :
             list);
}
*/
/*
NODEPTR prevnext_list(list)
NODEPTR list;
{
    return ((is_prev1 (node_lst_size(list),1) || is_prev1 (node_lst_base(list),1)) ?
            list_node (node_lst_blk(list),
                       prev_loc1 (list_size (list),1),
                       prev_loc1 (list_base (list)),1) :
             list);
}
*/
/*
NODEPTR prevprev_list(list)
NODEPTR list;
{
    return ((is_prevprev (node_lst_size(list)) || is_prevprev (node_lst_base(list))) ?
            list_node (node_lst_blk(list),
                        prevprev_loc (list_size (list)),
                        prevprev_loc (list_base (list))) :
             list);
}
*/

NODEPTR list_size1(list)
NODEPTR list;
{
    NODEPTR loc = node_lst_size(list);
    dereference (loc);
    return (loc);
}

NODEPTR list_base1(list)
NODEPTR list;
{
    NODEPTR loc = node_lst_base(list);
    dereference (loc);
    return (loc);
}

/*
NODEPTR put_mem (loc,val)
register NODEPTR loc;
NODEPTR val;
{
    NODEPTR adj_val, ploc, pnloc, pploc;
    dereference (loc);
    if (DebugLevel>2) {
        printf("***put_mem***: %s\n",is_next (loc)? "is_next" : is_prev (loc)? "is_prev" : is_prevnext(loc)? "is_prevnext": is_prevprev(loc)? "is_prevprev": "is_curr");
        show_mem_all(loc);
    }
    ploc = prev_loc(loc);
    pnloc = prevnext_loc(loc);
    pploc = prevprev_loc(loc);
    if (DebugLevel>5) {
        if (node_mem_isundefined(loc) == 1)
            printf ("changing variable from undefined to defined\n");
    }
    if (is_prevprev (loc)) {
        if (mem_defined(pploc)) {
            if (!equal_val (val, get_mem (pploc))) {
                error_break0 ("attempt to overwrite variable in the previous state");
            }
        }
        else
          error_break0 ("attempt to overwrite variable in the previous state");
    }
    else {
        if (is_next (loc)) {
            if (mem_defined (curr_loc (loc))) {
                if (val_defined2(get_mem(curr_loc (loc)))){
                    node_next(ploc) = copy_val2(get_mem(curr_loc (loc)));
                    node_mem_time(ploc) = Clock;
                }
                else
                    node_mem_time(ploc) = UNDEF;
            }
            else {
                node_mem_time(ploc) = UNDEF;
            }
            node_next(pnloc) = next_val (val);
            node_mem_time(pnloc) = Clock+1;
            node_next(loc) = adj_val = next_val (val);
            node_mem_time(loc) = Clock + 1;
            if (node_mem_isstatic(loc)) {
                node_next(loc = curr_loc (loc)) = curr_val (val);
                node_mem_time(loc) = Clock;
            }    
        }
        else {
            if (mem_defined (next_loc(loc))) {
                if (val_defined2(get_mem(next_loc(loc)))) {
                    node_next(pnloc) = copy_val2(get_mem(next_loc(loc)));
                    node_mem_time(pnloc) = (Clock == 0)? UNDEF : Clock -1 ;
                }
                else
                    node_mem_time(pnloc) =  UNDEF;
            }
            else {
                node_mem_time(pnloc) =  UNDEF;
            }
            node_next(ploc) = curr_val (val);
            node_mem_time(ploc) = Clock;
            node_next(loc) = adj_val = curr_val (val);
            node_mem_time(loc) = Clock;
            if (node_mem_isstatic(loc)) {
                node_next(loc = next_loc (loc)) = next_val (val);
                node_mem_time(loc) = Clock + 1;
            }
        }
    }
    if (DebugLevel>2) {
        printf("-------------------\n");
        show_mem_all(loc);
        
    }
    return (adj_val);
}
*/

NODEPTR put_mem1 (loc,val)
register NODEPTR loc;
NODEPTR val;
{
    NODEPTR adj_val, ploc, pnloc, pploc;
    unsigned int i;
    dereference (loc);
    if (DebugLevel>2) {
        printf("***put_mem***: \n");
        show_mem_all1(loc);
    }
        /*if (!val_defined(val)){
        printf("*****putmem***this should not happen\n");
        output_expn(val, -1);
        show_mem_all1(loc);
        }*/
    ploc = prev_loc1(loc,0);
    pnloc = prev_loc1(loc,1);
    pploc = prev_loc1(loc,node_mem_history(loc)+1);
    if (DebugLevel>5) {
        if (node_mem_isundefined(loc)==1)
            printf ("changing variable from undefined to defined\n");
    }
    if (is_prev1 (loc,node_mem_history(loc)+1)) {
        if (mem_defined1(pploc)) {
            if (!equal_val (val, get_mem (pploc))) {
                error_break0 ("attempt to overwrite variable in the previous state");
            }
        }
        else
          error_break0 ("attempt to overwrite variable in the previous state");
    }
    else {
        if (is_next (loc)) {
            if (mem_defined1 (curr_loc (loc))) {
                if (val_defined2(get_mem(curr_loc (loc)))){
                    node_next(ploc) = copy_val2(get_mem(curr_loc (loc)));
                    node_mem_time(ploc) = Clock;
                }
                else
                    node_mem_time(ploc) = UNDEF;
            }
            else {
                node_mem_time(ploc) = UNDEF;
            }
            node_next(pnloc) = next_val (val);
            node_mem_time(pnloc) = Clock+1;
            node_next(loc) = adj_val = next_val (val);
            node_mem_time(loc) = Clock + 1;
            if (node_mem_isstatic(loc)) {
                node_next(loc = curr_loc (loc)) = curr_val (val);
                node_mem_time(loc) = Clock;
                node_next(ploc) = curr_val (val);
                node_mem_time(ploc) = Clock;
                for (i=2;i<=1+node_mem_history(loc);i = i+1) {
                    node_next(prev_loc1(loc,i)) = curr_val (val);
                    node_mem_time(prev_loc1(loc,i)) = Clock + i;
                }
            }    
        }
        else {
            if (mem_defined1 (next_loc(loc))) {
                if (val_defined2(get_mem(next_loc(loc)))) {
                    node_next(pnloc) = copy_val2(get_mem(next_loc(loc)));
                    node_mem_time(pnloc) = (Clock == 0)? UNDEF : Clock -1 ;
                }
                else
                    node_mem_time(pnloc) =  UNDEF;
            }
            else {
                node_mem_time(pnloc) =  UNDEF;
            }
            node_next(ploc) = adj_val = curr_val (val);
            node_mem_time(ploc) = Clock;
            node_next(loc) = adj_val = curr_val (val);
            node_mem_time(loc) = Clock;
            if (node_mem_isstatic(loc)) {
                node_next(loc = next_loc (loc)) = next_val (val);
                node_mem_time(loc) = Clock + 1;
                node_next(pnloc) = next_val (val);
                node_mem_time(pnloc) = Clock + 1;
                for (i=2;i<=1+node_mem_history(loc);i = i+1) {
                    node_next(prev_loc1(loc,i)) = next_val (val);
                    node_mem_time(prev_loc1(loc,i)) = Clock + i;
                }
            }
        }
    }
    if (DebugLevel>2) {
        printf("-------------------\n");
        show_mem_all1(loc);
        
    }
    return (adj_val);
}

int is_static_mem(loc)
register NODEPTR loc;
{
    dereference (loc);
    return (node_mem_isstatic(loc));
}

void make_static_mem(loc)
register NODEPTR loc;
{
    unsigned int i;
    
    dereference (loc);
    node_mem_isstatic(next_loc (loc)) = node_mem_isstatic(curr_loc (loc)) = T;
    for (i=0; i <= 1+node_mem_history(loc); i = i + 1) {
        node_mem_isstatic(prev_loc1(loc,i)) = T;
    }
}

void make_indirect_mem(loc,val)
NODEPTR loc;
register NODEPTR val;
{
    NODEPTR cloc, nloc, ploc;
    dereference (val);
    cloc = curr_loc (loc);
    nloc = next_loc (loc);
    node_next(cloc) = val;
    node_next(nloc) = is_curr (val) ? next_loc (val) : curr_loc (val);
    node_mem_isindirect(cloc) = node_mem_isindirect(nloc) = T;
    
    
}

void erase_mem(loc)
register NODEPTR loc;
{
    
    dereference(loc);
    if (DebugLevel>2) {
        printf("***erase_mem***pre, Clock=%4d\n",Clock);
        show_mem_all1(loc);
    }
    node_mem_time(loc) = UNDEF;
    node_next(loc) = NULL;
    node_mem_isundefined(loc) = 0;
    node_mem_hasdefault(loc) = 0;
    if (DebugLevel>2) {
        printf("***erase_mem***post\n");
        show_mem_all1(loc);
    }
}



/******************************\
*                              *
*  Routines for the empty-flag *
*                              *
\******************************/

int empty_defined()
{
    return (mem_defined1 (EmptyLoc));
}

void put_empty(val)
NODEPTR val;
{
    if (mem_defined1 (EmptyLoc)) {
        assert0 (equal_val (val, get_mem (EmptyLoc)),
                "attempt to re-assign interval length");
    }
    else
        put_mem1 (EmptyLoc, val);
}

NODEPTR get_empty()
{
    return (node_next(EmptyLoc));
}



/*************************\
*                         *
*  Environment handling   *
*                         *
\*************************/

void push_env(bnd)
NODEPTR bnd;
{
    if (EnvTop >= MAXENVSTACKSIZE)
        error_break0("environment stack full");
    else {
        EnvTop += 1;
        EnvStack[EnvTop].env = node_bnd_env(bnd);
        EnvStack[EnvTop].env_base = node_blk_base(node_bnd_blk(bnd));
    }
}

void pop_env()
{
    EnvTop -= 1;
}

void push_display(bnd)
NODEPTR bnd;
{
    register unsigned l = node_env_level(node_bnd_env(bnd));
    unsigned newenvbase;
    if (EnvTop + l + 1 >= MAXENVSTACKSIZE)
        error_break0 ("environment stack full");
    else {
	newenvbase = (EnvTop += 1);
        for (; l > 0; l -= 1) {
            EnvStack[EnvTop].env = EnvStack[EnvBase].env;
            EnvStack[EnvTop].env_base = EnvStack[EnvBase].env_base;
            EnvTop += 1;
            EnvBase += 1;
	}
        EnvStack[EnvTop].env_base = node_blk_base(node_bnd_blk(bnd));
        EnvStack[EnvTop].env = node_bnd_env(bnd);
        EnvBase = newenvbase;
    }
}

void pop_display()
{
    EnvTop = --EnvBase;
    EnvBase = EnvBase - (node_env_level(EnvStack[EnvTop].env));
}


